import json
import pytest
import websockets


@pytest.mark.asyncio
@pytest.mark.parametrize("server", ["asgi", "rsgi"], indirect=True)
@pytest.mark.parametrize("threading_mode", ["runtime", "workers"])
async def test_messages(server, threading_mode):
    async with server(threading_mode) as port:
        async with websockets.connect(f"ws://localhost:{port}/ws_echo") as ws:
            await ws.send("foo")
            res_text = await ws.recv()
            await ws.send(b"foo")
            res_bytes = await ws.recv()

    assert res_text == "foo"
    assert res_bytes == b"foo"


@pytest.mark.asyncio
@pytest.mark.parametrize("server", ["asgi", "rsgi"], indirect=True)
@pytest.mark.parametrize("threading_mode", ["runtime", "workers"])
async def test_reject(server, threading_mode):
    async with server(threading_mode) as port:
        with pytest.raises(websockets.InvalidStatusCode) as exc:
            async with websockets.connect(f"ws://localhost:{port}/ws_reject") as ws:
                pass

    assert exc.value.status_code == 403


@pytest.mark.asyncio
@pytest.mark.skip
@pytest.mark.parametrize("server", ["asgi", "rsgi"], indirect=True)
@pytest.mark.parametrize("threading_mode", ["runtime", "workers"])
async def test_disconnect(server, threading_mode):
    count = 0
    async with server(threading_mode) as port:
        async with websockets.connect(f"ws://localhost:{port}/ws_push") as ws:
            while count < 3:
                await ws.recv()
                count += 1

    assert count == 3


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "threading_mode",
    [
        "runtime",
        "workers"
    ]
)
async def test_asgi_scope(asgi_server, threading_mode):
    async with asgi_server(threading_mode) as port:
        async with websockets.connect(f"ws://localhost:{port}/ws_info?test=true") as ws:
            res = await ws.recv()

    data = json.loads(res)
    assert data['asgi'] == {
        'version': '3.0',
        'spec_version': '2.3'
    }
    assert data['type'] == "websocket"
    assert data['http_version'] == '1.1'
    assert data['scheme'] == 'ws'
    assert data['path'] == '/ws_info'
    assert data['query_string'] == 'test=true'
    assert data['headers']['host'] == f'localhost:{port}'


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "threading_mode",
    [
        "runtime",
        "workers"
    ]
)
async def test_rsgi_scope(rsgi_server, threading_mode):
    async with rsgi_server(threading_mode) as port:
        async with websockets.connect(f"ws://localhost:{port}/ws_info?test=true") as ws:
            res = await ws.recv()

    data = json.loads(res)
    assert data['proto'] == 'ws'
    assert data['http_version'] == '1.1'
    assert data['rsgi_version'] == '1.0'
    assert data['scheme'] == 'http'
    assert data['method'] == "GET"
    assert data['path'] == '/ws_info'
    assert data['query_string'] == 'test=true'
    assert data['headers']['host'] == f'localhost:{port}'



from typing import Any, Dict, List, Tuple, Optional

from ._types import WebsocketMessage


class ASGIScope:
    client_ip: str
    client_port: int
    server_ip: str
    server_port: int
    http_version: str
    method: str
    path: str
    proto: str
    query_string: str
    scheme: str

    @property
    def headers(self) -> List[Tuple[bytes, bytes]]: ...


class RSGIHeaders:
    def __contains__(self, key: str) -> bool: ...
    def keys(self) -> List[str]: ...
    def values(self) -> List[str]: ...
    def items(self) -> List[Tuple[str]]: ...
    def get(self, key: str, default: Any = None) -> Any: ...


class RSGIScope:
    proto: str
    http_version: str
    rsgi_version: str
    server: str
    client: str
    scheme: str
    method: str
    path: str
    query_string: str

    @property
    def headers(self) -> RSGIHeaders: ...


class RSGIHTTPProtocol:
    async def __call__(self) -> bytes: ...
    def response_empty(self, status: int, headers: List[Tuple[str, str]]): ...
    def response_str(self, status: int, headers: List[Tuple[str, str]], body: str): ...
    def response_bytes(self, status: int, headers: List[Tuple[str, str]], body: bytes): ...
    def response_file(self, status: int, headers: List[Tuple[str, str]], file: str): ...


class RSGIWebsocketTransport:
    async def receive(self) -> WebsocketMessage: ...
    async def send_bytes(self, data: bytes): ...
    async def send_str(self, data: str): ...


class RSGIWebsocketProtocol:
    async def accept(self) -> RSGIWebsocketTransport: ...
    def close(self, status: Optional[int]) -> Tuple[int, bool]: ...


class RSGIProtocolError(RuntimeError):
    ...


class RSGIProtocolClosed(RuntimeError):
    ...


class WSGIScope:
    server: str
    client: str
    scheme: str
    method: str
    path: str
    query_string: str
    headers: Dict[str, str]
    body: bytes



from enum import Enum


class Interfaces(str, Enum):
    ASGI = "asgi"
    RSGI = "rsgi"
    WSGI = "wsgi"


class HTTPModes(str, Enum):
    auto = "auto"
    http1 = "1"
    http2 = "2"


class ThreadModes(str, Enum):
    runtime = "runtime"
    workers = "workers"


class Loops(str, Enum):
    auto = "auto"
    asyncio = "asyncio"
    uvloop = "uvloop"



import asyncio
import os
import socket

from contextlib import asynccontextmanager, closing
from functools import partial
from pathlib import Path

import pytest


@asynccontextmanager
async def _server(interface, port, threading_mode, tls=False):
    certs_path = Path.cwd() / "tests" / "fixtures" / "tls"
    tls_opts = (
        f"--ssl-certificate {certs_path / 'cert.pem'} "
        f"--ssl-keyfile {certs_path / 'key.pem'} "
    ) if tls else ""
    proc = await asyncio.create_subprocess_shell(
        "".join([
            f"granian --interface {interface} --port {port} ",
            f"--threads 1 --threading-mode {threading_mode} ",
            tls_opts,
            f"tests.apps.{interface}:app"
        ]),
        env=dict(os.environ)
    )
    await asyncio.sleep(1)
    try:
        yield port
    finally:
        proc.terminate()
        await proc.wait()


@pytest.fixture(scope="function")
def server_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind(('localhost', 0))
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return sock.getsockname()[1]


@pytest.fixture(scope="function")
def asgi_server(server_port):
    return partial(_server, "asgi", server_port)


@pytest.fixture(scope="function")
def rsgi_server(server_port):
    return partial(_server, "rsgi", server_port)


@pytest.fixture(scope="function")
def wsgi_server(server_port):
    return partial(_server, "wsgi", server_port)


@pytest.fixture(scope="function")
def server(server_port, request):
    return partial(_server, request.param, server_port)


@pytest.fixture(scope="function")
def server_tls(server_port, request):
    return partial(_server, request.param, server_port, tls=True)



import json

PLAINTEXT_RESPONSE = {
    'type': 'http.response.start',
    'status': 200,
    'headers': [
        [b'content-type', b'text/plain; charset=utf-8'],
    ]
}
JSON_RESPONSE = {
    'type': 'http.response.start',
    'status': 200,
    'headers': [
        [b'content-type', b'application/json'],
    ]
}


async def info(scope, receive, send):
    await send(JSON_RESPONSE)
    await send({
        'type': 'http.response.body',
        'body': json.dumps({
            'type': scope['type'],
            'asgi': scope['asgi'],
            'http_version': scope['http_version'],
            'scheme': scope['scheme'],
            'method': scope['method'],
            'path': scope['path'],
            'query_string': scope['query_string'].decode("latin-1"),
            'headers': {
                k.decode("utf8"): v.decode("utf8")
                for k, v in scope['headers']
            }
        }).encode("utf8"),
        'more_body': False
    })


async def echo(scope, receive, send):
    await send(PLAINTEXT_RESPONSE)
    msg = await receive()
    await send({
        'type': 'http.response.body',
        'body': msg['body'],
        'more_body': False
    })


async def ws_reject(scope, receive, send):
    return


async def ws_info(scope, receive, send):
    await send({'type': 'websocket.accept'})
    await send({
        'type': 'websocket.send',
        'text': json.dumps({
            'type': scope['type'],
            'asgi': scope['asgi'],
            'http_version': scope['http_version'],
            'scheme': scope['scheme'],
            'path': scope['path'],
            'query_string': scope['query_string'].decode("latin-1"),
            'headers': {
                k.decode("utf8"): v.decode("utf8")
                for k, v in scope['headers']
            }
        })
    })
    await send({'type': 'websocket.close'})


async def ws_echo(scope, receive, send):
    await send({'type': 'websocket.accept'})

    while True:
        msg = await receive()
        if msg['type'] == 'websocket.connect':
            continue
        if msg['type'] == 'websocket.disconnect':
            break
        rv = {'type': 'websocket.send', 'bytes': None, 'text': None}
        key = 'text' if 'text' in msg else 'bytes'
        rv[key] = msg[key]
        await send(rv)


async def ws_push(scope, receive, send):
    await send({'type': 'websocket.accept'})

    try:
        while True:
            await send({
                'type': 'websocket.send',
                'text': 'ping'
            })
    except Exception:
        pass


async def err_app(scope, receive, send):
    1 / 0


async def err_proto(scope, receive, send):
    await send({'type': 'wrong.msg'})


def app(scope, receive, send):
    return {
        "/info": info,
        "/echo": echo,
        "/ws_reject": ws_reject,
        "/ws_info": ws_info,
        "/ws_echo": ws_echo,
        "/ws_push": ws_push,
        "/err_app": err_app,
        "/err_proto": err_proto
    }[scope['path']](scope, receive, send)



from pathlib import Path
from typing import Optional

import typer

from .__version__ import __version__
from .constants import Interfaces, HTTPModes, Loops, ThreadModes
from .log import LogLevels
from .server import Granian


cli = typer.Typer(name="granian", context_settings={"ignore_unknown_options": True})


def version_callback(value: bool):
    if value:
        typer.echo(f"{cli.info.name} {__version__}")
        raise typer.Exit()


@cli.command()
def main(
    app: str = typer.Argument(..., help="Application target to serve."),
    host: str = typer.Option("127.0.0.1", help="Host address to bind to."),
    port: int = typer.Option(8000, help="Port to bind to."),
    interface: Interfaces = typer.Option(
        Interfaces.RSGI.value,
        help="Application interface type."
    ),
    http: HTTPModes = typer.Option(
        HTTPModes.auto.value,
        help="HTTP version."
    ),
    websockets: bool = typer.Option(
        True,
        "--ws/--no-ws",
        help="Enable websockets handling",
        show_default="enabled"
    ),
    workers: int = typer.Option(1, min=1, help="Number of worker processes."),
    threads: int = typer.Option(1, min=1, help="Number of threads."),
    threading_mode: ThreadModes = typer.Option(
        ThreadModes.workers.value,
        help="Threading mode to use."
    ),
    loop: Loops = typer.Option(Loops.auto.value, help="Event loop implementation"),
    backlog: int = typer.Option(
        1024,
        min=128,
        help="Maximum number of connections to hold in backlog."
    ),
    log_level: LogLevels = typer.Option(
        LogLevels.info.value,
        help="Log level",
        case_sensitive=False
    ),
    ssl_keyfile: Optional[Path] = typer.Option(
        None,
        help="SSL key file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True
    ),
    ssl_certificate: Optional[Path] = typer.Option(
        None,
        help="SSL certificate file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True
    ),
    _: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Shows the version and exit."
    )
):
    Granian(
        app,
        address=host,
        port=port,
        interface=interface,
        workers=workers,
        threads=threads,
        pthreads=threads,
        threading_mode=threading_mode,
        loop=loop,
        http=http,
        websockets=websockets,
        backlog=backlog,
        log_level=log_level,
        ssl_cert=ssl_certificate,
        ssl_key=ssl_keyfile
    ).serve()



import httpx
import json
import pathlib
import pytest
import ssl
import websockets


@pytest.mark.asyncio
@pytest.mark.parametrize("server_tls", ["asgi", "rsgi"], indirect=True)
@pytest.mark.parametrize("threading_mode", ["runtime", "workers"])
async def test_http_scope(server_tls, threading_mode):
    async with server_tls(threading_mode) as port:
        res = httpx.get(
            f"https://localhost:{port}/info?test=true",
            verify=False
        )

    assert res.status_code == 200
    data = res.json()
    assert data['scheme'] == 'https'


@pytest.mark.asyncio
@pytest.mark.parametrize("threading_mode", ["runtime", "workers"])
async def test_asgi_ws_scope(asgi_server, threading_mode):
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    localhost_pem = pathlib.Path.cwd() / "tests" / "fixtures" / "tls" / "cert.pem"
    ssl_context.load_verify_locations(localhost_pem)

    async with asgi_server(threading_mode, tls=True) as port:
        async with websockets.connect(
            f"wss://localhost:{port}/ws_info?test=true",
            ssl=ssl_context
        ) as ws:
            res = await ws.recv()

    data = json.loads(res)
    assert data['scheme'] == 'wss'


@pytest.mark.asyncio
@pytest.mark.parametrize("threading_mode", ["runtime", "workers"])
async def test_rsgi_ws_scope(rsgi_server, threading_mode):
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    localhost_pem = pathlib.Path.cwd() / "tests" / "fixtures" / "tls" / "cert.pem"
    ssl_context.load_verify_locations(localhost_pem)

    async with rsgi_server(threading_mode, tls=True) as port:
        async with websockets.connect(
            f"wss://localhost:{port}/ws_info?test=true",
            ssl=ssl_context
        ) as ws:
            res = await ws.recv()

    data = json.loads(res)
    assert data['scheme'] == 'https'



import httpx
import pytest


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "threading_mode",
    [
        "runtime",
        "workers"
    ]
)
async def test_scope(wsgi_server, threading_mode):
    payload = "body_payload"
    async with wsgi_server(threading_mode) as port:
        res = httpx.post(f"http://localhost:{port}/info?test=true", data=payload)

    assert res.status_code == 200
    assert res.headers["content-type"] == "application/json"

    data = res.json()
    assert data['scheme'] == 'http'
    assert data['method'] == "POST"
    assert data['path'] == '/info'
    assert data['query_string'] == 'test=true'
    assert data['headers']['HTTP_HOST'] == f'localhost:{port}'
    assert data['content_length'] == str(len(payload))


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "threading_mode",
    [
        "runtime",
        "workers"
    ]
)
async def test_body(wsgi_server, threading_mode):
    async with wsgi_server(threading_mode) as port:
        res = httpx.post(f"http://localhost:{port}/echo", data="test")

    assert res.status_code == 200
    assert res.text == "test"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "threading_mode",
    [
        "runtime",
        "workers"
    ]
)
async def test_app_error(wsgi_server, threading_mode):
    async with wsgi_server(threading_mode) as port:
        res = httpx.get(f"http://localhost:{port}/err_app")

    assert res.status_code == 500



import asyncio

from functools import wraps

from ._futures import future_with_watcher
from ._granian import ASGIScope as Scope


class LifespanProtocol:
    error_transition = 'Invalid lifespan state transition'

    def __init__(self, callable):
        self.callable = callable
        self.event_queue = asyncio.Queue()
        self.event_startup = asyncio.Event()
        self.event_shutdown = asyncio.Event()
        self.unsupported = False
        self.errored = False
        self.failure_startup = False
        self.failure_shutdown = False
        self.interrupt = False

    async def handle(self):
        try:
            await self.callable(
                {
                    "type": "lifespan",
                    "asgi": {"version": "3.0", "spec_version": "2.3"}
                },
                self.receive,
                self.send
            )
        except Exception:
            self.errored = True
            if self.failure_startup or self.failure_shutdown:
                return
            self.unsupported = True
            # self.logger.error(msg, exc_info=exc)
        finally:
            self.event_startup.set()
            self.event_shutdown.set()

    async def startup(self):
        loop = asyncio.get_event_loop()
        handler_task = loop.create_task(self.handle())

        await self.event_queue.put({"type": "lifespan.startup"})
        await self.event_startup.wait()

        if self.failure_startup or (self.errored and not self.unsupported):
            self.interrupt = True

    async def shutdown(self):
        if self.errored:
            return

        await self.event_queue.put({"type": "lifespan.shutdown"})
        await self.event_shutdown.wait()

        if self.failure_shutdown or (self.errored and not self.unsupported):
            self.interrupt = True

    async def receive(self):
        return await self.event_queue.get()

    def _handle_startup_complete(self, message):
        assert not self.event_startup.is_set(), self.error_transition
        assert not self.event_shutdown.is_set(), self.error_transition
        self.event_startup.set()

    def _handle_startup_failed(self, message):
        assert not self.event_startup.is_set(), self.error_transition
        assert not self.event_shutdown.is_set(), self.error_transition
        self.event_startup.set()
        self.failure_startup = True
        # if message.get("message"):
        #     self.logger.error(message["message"])

    def _handle_shutdown_complete(self, message):
        assert self.event_startup.is_set(), self.error_transition
        assert not self.event_shutdown.is_set(), self.error_transition
        self.event_shutdown.set()

    def _handle_shutdown_failed(self, message):
        assert self.event_startup.is_set(), self.error_transition
        assert not self.event_shutdown.is_set(), self.error_transition
        self.event_shutdown.set()
        self.failure_shutdown = True
        # if message.get("message"):
        #     self.logger.error(message["message"])

    _event_handlers = {
        "lifespan.startup.complete": _handle_startup_complete,
        "lifespan.startup.failed": _handle_startup_failed,
        "lifespan.shutdown.complete": _handle_shutdown_complete,
        "lifespan.shutdown.failed": _handle_shutdown_failed
    }

    async def send(self, message):
        handler = self._event_handlers[message["type"]]
        handler(self, message)


def _noop_wrapper(proto):
    return proto


def _send_http_wrapper(proto):
    @wraps(proto)
    async def send(data):
        proto(data)
    return send


_send_wrappers = {
    "http": _send_http_wrapper,
    "websocket": _noop_wrapper
}


def _callback_wrapper(callback):
    @wraps(callback)
    def wrapper(watcher, scope: Scope):
        coro = callback(
            {
                "type": scope.proto,
                "asgi": {
                    "version": "3.0",
                    "spec_version": "2.3"
                },
                "http_version": scope.http_version,
                "server": (scope.server_ip, scope.server_port),
                "client": (scope.client_ip, scope.client_port),
                "scheme": scope.scheme,
                "method": scope.method,
                "root_path": "",
                "path": scope.path,
                "raw_path": scope.path.encode("ascii"),
                "query_string": scope.query_string.encode('latin-1'),
                "headers": scope.headers,
                "extensions": {}
            },
            watcher.proto.receive,
            _send_wrappers[scope.proto](watcher.proto.send)
        )
        watcher.event_loop.call_soon_threadsafe(
            future_with_watcher,
            coro,
            watcher,
            context=watcher.context
        )
    return wrapper



import json

from granian.rsgi import (
    HTTPProtocol,
    Scope,
    WebsocketMessageType,
    WebsocketProtocol
)


async def info(scope: Scope, protocol: HTTPProtocol):
    protocol.response_bytes(
        200,
        [('content-type', 'application/json')],
        json.dumps({
            'proto': scope.proto,
            'http_version': scope.http_version,
            'rsgi_version': scope.rsgi_version,
            'scheme': scope.scheme,
            'method': scope.method,
            'path': scope.path,
            'query_string': scope.query_string,
            'headers': {k: v for k, v in scope.headers.items()}
        }).encode("utf8")
    )


async def echo(_, protocol: HTTPProtocol):
    msg = await protocol()
    protocol.response_bytes(
        200,
        [('content-type', 'text/plain; charset=utf-8')],
        msg
    )


async def ws_reject(_, protocol: WebsocketProtocol):
    protocol.close(403)


async def ws_info(scope: Scope, protocol: WebsocketProtocol):
    trx = await protocol.accept()

    await trx.send_str(json.dumps({
        'proto': scope.proto,
        'http_version': scope.http_version,
        'rsgi_version': scope.rsgi_version,
        'scheme': scope.scheme,
        'method': scope.method,
        'path': scope.path,
        'query_string': scope.query_string,
        'headers': {k: v for k, v in scope.headers.items()}
    }))
    while True:
        message = await trx.receive()
        if message.kind == WebsocketMessageType.close:
            break

    protocol.close()


async def ws_echo(_, protocol: WebsocketProtocol):
    trx = await protocol.accept()

    while True:
        message = await trx.receive()
        if message.kind == WebsocketMessageType.close:
            break
        elif message.kind == WebsocketMessageType.bytes:
            await trx.send_bytes(message.data)
        else:
            await trx.send_str(message.data)

    protocol.close()


async def ws_push(_, protocol: WebsocketProtocol):
    trx = await protocol.accept()

    try:
        while True:
            await trx.send_str("ping")
    except Exception:
        pass

    protocol.close()


async def err_app(scope: Scope, protocol: HTTPProtocol):
    1 / 0


def app(scope, protocol):
    return {
        "/info": info,
        "/echo": echo,
        "/ws_reject": ws_reject,
        "/ws_info": ws_info,
        "/ws_echo": ws_echo,
        "/ws_push": ws_push,
        "/err_app": err_app
    }[scope.path](scope, protocol)



from enum import Enum
from functools import wraps
from typing import Union

from ._futures import future_with_watcher
from ._granian import (
    RSGIHTTPProtocol as HTTPProtocol,
    RSGIWebsocketProtocol as WebsocketProtocol,
    RSGIHeaders as Headers,
    RSGIScope as Scope,
    RSGIProtocolError as ProtocolError,
    RSGIProtocolClosed as ProtocolClosed
)


class WebsocketMessageType(int, Enum):
    close = 0
    bytes = 1
    string = 2


class WebsocketMessage:
    kind: WebsocketMessageType
    data: Union[bytes, str]


def _callback_wrapper(callback):
    @wraps(callback)
    def wrapper(watcher, scope: Scope):
        watcher.event_loop.call_soon_threadsafe(
            future_with_watcher,
            callback(scope, watcher.proto),
            watcher,
            context=watcher.context
        )
    return wrapper



import os
import sys

from functools import wraps
from typing import Any, List, Tuple

from ._granian import WSGIScope as Scope


class Response:
    __slots__ = ['status', 'headers']

    def __init__(self):
        self.status = 200
        self.headers = []


def _callback_wrapper(callback):
    basic_env = dict(os.environ)
    basic_env.update({
        'GATEWAY_INTERFACE': 'CGI/1.1',
        'SCRIPT_NAME': '',
        'SERVER_PROTOCOL': 'HTTP/1.1',
        'SERVER_SOFTWARE': 'Granian',
        'wsgi.errors': sys.stderr,
        'wsgi.input_terminated': True,
        'wsgi.input': None,
        'wsgi.multiprocess': False,
        'wsgi.multithread': False,
        'wsgi.run_once': False,
        'wsgi.version': (1, 0)
    })

    @wraps(callback)
    def wrapper(scope: Scope) -> Tuple[int, List[Tuple[str, str]], bytes]:
        addr_server = scope.server.split(":")
        environ = {
            **basic_env,
            **scope.headers,
            'SERVER_NAME': addr_server[0],
            'SERVER_PORT': str(addr_server[1]),
            'REQUEST_METHOD': scope.method,
            'PATH_INFO': scope.path,
            'QUERY_STRING': scope.query_string,
            'REMOTE_ADDR': scope.client,
            'wsgi.url_scheme': scope.scheme,
            'wsgi.input': scope.input()
        }
        if 'HTTP_CONTENT_TYPE' in environ:
            environ['CONTENT_TYPE'] = environ.pop('HTTP_CONTENT_TYPE')
        if 'HTTP_CONTENT_LENGTH' in environ:
            environ['CONTENT_LENGTH'] = environ.pop('HTTP_CONTENT_LENGTH')

        resp = Response()

        def start_response(
            status: str,
            headers: List[Tuple[str, str]],
            exc_info: Any = None
        ):
            resp.status = int(status.split(' ', 1)[0])
            resp.headers = headers

        rv = callback(environ, start_response)

        try:
            body = b"".join(rv)
        finally:
            if hasattr(rv, "close"):
                rv.close()

        return (resp.status, resp.headers, body)

    return wrapper



import contextvars
import multiprocessing
import signal
import socket
import ssl
import threading

from functools import partial
from pathlib import Path
from typing import List, Optional

from ._granian import ASGIWorker, RSGIWorker, WSGIWorker
from ._internal import load_target
from .asgi import LifespanProtocol, _callback_wrapper as _asgi_call_wrap
from .constants import Interfaces, HTTPModes, Loops, ThreadModes
from .log import LogLevels, configure_logging, logger
from .net import SocketHolder
from .rsgi import _callback_wrapper as _rsgi_call_wrap
from .wsgi import _callback_wrapper as _wsgi_call_wrap

multiprocessing.allow_connection_pickling()


class Granian:
    SIGNALS = {signal.SIGINT, signal.SIGTERM}

    def __init__(
        self,
        target: str,
        address: str = "127.0.0.1",
        port: int = 8000,
        interface: Interfaces = Interfaces.RSGI,
        workers: int = 1,
        threads: int = 1,
        pthreads: int = 1,
        threading_mode: ThreadModes = ThreadModes.workers,
        loop: Loops = Loops.auto,
        http: HTTPModes = HTTPModes.auto,
        websockets: bool = True,
        backlog: int = 1024,
        http1_buffer_size: int = 65535,
        log_level: LogLevels = LogLevels.info,
        ssl_cert: Optional[Path] = None,
        ssl_key: Optional[Path] = None
    ):
        self.target = target
        self.bind_addr = address
        self.bind_port = port
        self.interface = interface
        self.workers = max(1, workers)
        self.threads = max(1, threads)
        self.pthreads = max(1, pthreads)
        self.threading_mode = threading_mode
        self.loop = loop
        self.http = http
        self.websockets = websockets
        self.backlog = max(128, backlog)
        self.http1_buffer_size = http1_buffer_size
        self.log_level = log_level
        configure_logging(self.log_level)
        self.build_ssl_context(ssl_cert, ssl_key)
        self._shd = None
        self._sfd = None
        self.procs: List[multiprocessing.Process] = []
        self.exit_event = threading.Event()

    def build_ssl_context(
        self,
        cert: Optional[Path],
        key: Optional[Path]
    ):
        if not (cert and key):
            self.ssl_ctx = (False, None, None)
            return
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        ctx.load_cert_chain(cert, key, None)
        self.ssl_enabled = True
        # with cert.open("rb") as f:
        #     cert_contents = f.read()
        # with key.open("rb") as f:
        #     key_contents = f.read()
        self.ssl_ctx = (True, str(cert.resolve()), str(key.resolve()))

    @staticmethod
    def _spawn_asgi_worker(
        worker_id,
        callback_loader,
        socket,
        loop_impl,
        threads,
        pthreads,
        threading_mode,
        http_mode,
        http1_buffer_size,
        websockets,
        log_level,
        ssl_ctx
    ):
        from granian._loops import loops, set_loop_signals

        configure_logging(log_level)
        loop = loops.get(loop_impl)
        sfd = socket.fileno()
        callback = callback_loader()
        lifespan_handler = LifespanProtocol(callback)

        loop.run_until_complete(lifespan_handler.startup())
        if lifespan_handler.interrupt:
            return

        shutdown_event = set_loop_signals(loop, [signal.SIGTERM, signal.SIGINT])

        worker = ASGIWorker(
            worker_id,
            sfd,
            threads,
            pthreads,
            http_mode,
            http1_buffer_size,
            websockets,
            *ssl_ctx
        )
        serve = getattr(worker, {
            ThreadModes.runtime: "serve_rth",
            ThreadModes.workers: "serve_wth"
        }[threading_mode])
        serve(
            _asgi_call_wrap(callback),
            loop,
            contextvars.copy_context(),
            shutdown_event.wait()
        )
        loop.run_until_complete(lifespan_handler.shutdown())


    @staticmethod
    def _spawn_rsgi_worker(
        worker_id,
        callback_loader,
        socket,
        loop_impl,
        threads,
        pthreads,
        threading_mode,
        http_mode,
        http1_buffer_size,
        websockets,
        log_level,
        ssl_ctx
    ):
        from granian._loops import loops, set_loop_signals

        configure_logging(log_level)
        loop = loops.get(loop_impl)
        sfd = socket.fileno()
        target = callback_loader()
        callback = (
            getattr(target, '__rsgi__') if hasattr(target, '__rsgi__') else
            target
        )
        callback_init = (
            getattr(target, '__rsgi_init__') if hasattr(target, '__rsgi_init__') else
            lambda *args, **kwargs: None
        )

        shutdown_event = set_loop_signals(loop, [signal.SIGTERM, signal.SIGINT])
        callback_init(loop)

        worker = RSGIWorker(
            worker_id,
            sfd,
            threads,
            pthreads,
            http_mode,
            http1_buffer_size,
            websockets,
            *ssl_ctx
        )
        serve = getattr(worker, {
            ThreadModes.runtime: "serve_rth",
            ThreadModes.workers: "serve_wth"
        }[threading_mode])
        serve(
            _rsgi_call_wrap(callback),
            loop,
            contextvars.copy_context(),
            shutdown_event.wait()
        )

    @staticmethod
    def _spawn_wsgi_worker(
        worker_id,
        callback_loader,
        socket,
        loop_impl,
        threads,
        pthreads,
        threading_mode,
        http_mode,
        http1_buffer_size,
        websockets,
        log_level,
        ssl_ctx
    ):
        from granian._loops import loops, set_loop_signals

        configure_logging(log_level)
        loop = loops.get(loop_impl)
        sfd = socket.fileno()
        callback = callback_loader()

        shutdown_event = set_loop_signals(loop, [signal.SIGTERM, signal.SIGINT])

        worker = WSGIWorker(
            worker_id,
            sfd,
            threads,
            pthreads,
            http_mode,
            http1_buffer_size,
            *ssl_ctx
        )
        serve = getattr(worker, {
            ThreadModes.runtime: "serve_rth",
            ThreadModes.workers: "serve_wth"
        }[threading_mode])
        serve(
            _wsgi_call_wrap(callback),
            loop,
            contextvars.copy_context(),
            shutdown_event.wait()
        )

    def _init_shared_socket(self):
        self._shd = SocketHolder.from_address(
            self.bind_addr,
            self.bind_port,
            self.backlog
        )
        self._sfd = self._shd.get_fd()

    def signal_handler(self, *args, **kwargs):
        self.exit_event.set()

    def _spawn_proc(
        self,
        id,
        target,
        callback_loader,
        socket_loader
    ) -> multiprocessing.Process:
        return multiprocessing.get_context().Process(
            name="granian-worker",
            target=target,
            args=(
                id,
                callback_loader,
                socket_loader(),
                self.loop,
                self.threads,
                self.pthreads,
                self.threading_mode,
                self.http,
                self.http1_buffer_size,
                self.websockets,
                self.log_level,
                self.ssl_ctx
            )
        )

    def startup(self, spawn_target, target_loader):
        logger.info("Starting granian")

        for sig in self.SIGNALS:
            signal.signal(sig, self.signal_handler)

        self._init_shared_socket()
        sock = socket.socket(fileno=self._sfd)
        sock.set_inheritable(True)
        logger.info(f"Listening at: {self.bind_addr}:{self.bind_port}")

        def socket_loader():
            return sock

        for idx in range(self.workers):
            proc = self._spawn_proc(
                id=idx + 1,
                target=spawn_target,
                callback_loader=target_loader,
                socket_loader=socket_loader
            )
            proc.start()
            self.procs.append(proc)
            logger.info(f"Booting worker-{idx + 1} with pid: {proc.pid}")

    def shutdown(self):
        logger.info("Shutting down granian")
        for proc in self.procs:
            proc.terminate()
        for proc in self.procs:
            proc.join()

    def serve(self, spawn_target = None, target_loader = None):
        default_spawners = {
            Interfaces.ASGI: self._spawn_asgi_worker,
            Interfaces.RSGI: self._spawn_rsgi_worker,
            Interfaces.WSGI: self._spawn_wsgi_worker
        }
        target_loader = target_loader or load_target
        spawn_target = spawn_target or default_spawners[self.interface]

        self.startup(spawn_target, partial(target_loader, self.target))
        self.exit_event.wait()
        self.shutdown()



import json


def info(environ, protocol):
    protocol(
        "200 OK",
        [('content-type', 'application/json')]
    )
    return [json.dumps({
        'scheme': environ['wsgi.url_scheme'],
        'method': environ['REQUEST_METHOD'],
        'path': environ["PATH_INFO"],
        'query_string': environ["QUERY_STRING"],
        'content_length': environ['CONTENT_LENGTH'],
        'headers': {k: v for k, v in environ.items() if k.startswith("HTTP_")}
    }).encode("utf8")]


def echo(environ, protocol):
    protocol(
        '200 OK',
        [('content-type', 'text/plain; charset=utf-8')]
    )
    return [environ['wsgi.input'].read()]


def err_app(environ, protocol):
    1 / 0


def app(environ, protocol):
    return {
        "/info": info,
        "/echo": echo,
        "/err_app": err_app
    }[environ["PATH_INFO"]](environ, protocol)



async def future_watcher_wrapper(fut, watcher):
    try:
        await fut
    except Exception:
        watcher.err()
        raise
    watcher.done()


def future_with_watcher(coro, watcher):
    return watcher.event_loop.create_task(future_watcher_wrapper(coro, watcher))



import copy
import logging
import logging.config

from enum import Enum


class LogLevels(str, Enum):
    critical = "critical"
    error = "error"
    warning = "warning"
    warn = "warn"
    info = "info"
    debug = "debug"


log_levels_map = {
    LogLevels.critical: logging.CRITICAL,
    LogLevels.error: logging.ERROR,
    LogLevels.warning: logging.WARNING,
    LogLevels.warn: logging.WARN,
    LogLevels.info: logging.INFO,
    LogLevels.debug: logging.DEBUG
}

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "root": {"level": "INFO", "handlers": ["console"]},
    "formatters": {
        "generic": {
            "()": "logging.Formatter",
            "fmt": "[%(levelname)s] %(message)s",
            "datefmt": "[%Y-%m-%d %H:%M:%S %z]"
        }
    },
    "handlers": {
        "console": {
            "formatter": "generic",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        }
    }
}

logger = logging.getLogger()


def configure_logging(level: LogLevels):
    config = copy.deepcopy(LOGGING_CONFIG)
    config["root"]["level"] = log_levels_map[level]
    logging.config.dictConfig(config)



import httpx
import pytest


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "threading_mode",
    [
        "runtime",
        "workers"
    ]
)
async def test_scope(rsgi_server, threading_mode):
    async with rsgi_server(threading_mode) as port:
        res = httpx.get(f"http://localhost:{port}/info?test=true")

    assert res.status_code == 200
    assert res.headers["content-type"] == "application/json"

    data = res.json()
    assert data['proto'] == "http"
    assert data['http_version'] == '1.1'
    assert data['rsgi_version'] == '1.0'
    assert data['scheme'] == 'http'
    assert data['method'] == "GET"
    assert data['path'] == '/info'
    assert data['query_string'] == 'test=true'
    assert data['headers']['host'] == f'localhost:{port}'


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "threading_mode",
    [
        "runtime",
        "workers"
    ]
)
async def test_body(rsgi_server, threading_mode):
    async with rsgi_server(threading_mode) as port:
        res = httpx.post(f"http://localhost:{port}/echo", data="test")

    assert res.status_code == 200
    assert res.text == "test"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "threading_mode",
    [
        "runtime",
        "workers"
    ]
)
async def test_app_error(rsgi_server, threading_mode):
    async with rsgi_server(threading_mode) as port:
        res = httpx.get(f"http://localhost:{port}/err_app")

    assert res.status_code == 500



import asyncio
import os
import signal
import sys

from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple


class Registry:
    __slots__ = ["_data"]

    def __init__(self):
        self._data: Dict[str, Callable[..., Any]] = {}

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def keys(self) -> Iterable[str]:
        return self._data.keys()

    def register(self, key: str) -> Callable[[], Callable[..., Any]]:
        def wrap(builder: Callable[..., Any]) -> Callable[..., Any]:
            self._data[key] = builder
            return builder
        return wrap

    def get(self, key: str) -> Callable[..., Any]:
        try:
            return self._data[key]
        except KeyError:
            raise RuntimeError(f"'{key}' implementation not available.")



class BuilderRegistry(Registry):
    __slots__ = []

    def __init__(self):
        self._data: Dict[str, Tuple[Callable[..., Any], List[str]]] = {}

    def register(
        self,
        key: str,
        packages: Optional[List[str]] = None
    ) -> Callable[[], Callable[..., Any]]:
        packages = packages or []

        def wrap(builder: Callable[..., Any]) -> Callable[..., Any]:
            loaded_packages, implemented = {}, True
            try:
                for package in packages:
                    __import__(package)
                    loaded_packages[package] = sys.modules[package]
            except ImportError:
                implemented = False
            if implemented:
                self._data[key] = (builder, loaded_packages)
            return builder
        return wrap

    def get(self, key: str) -> Callable[..., Any]:
        try:
            builder, packages = self._data[key]
        except KeyError:
            raise RuntimeError(f"'{key}' implementation not available.")
        return builder(**packages)


loops = BuilderRegistry()


@loops.register('asyncio')
def build_asyncio_loop():
    loop = asyncio.new_event_loop() if os.name != 'nt' else asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)
    return loop


@loops.register('uvloop', packages=['uvloop'])
def build_uv_loop(uvloop):
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    return loop


@loops.register('auto')
def build_auto_loop():
    if 'uvloop' in loops:
        return loops.get('uvloop')
    return loops.get('asyncio')


def set_loop_signals(loop, signals):
    signal_event = asyncio.Event()

    def signal_handler(signum, frame):
        signal_event.set()

    try:
        for sigval in signals:
            loop.add_signal_handler(sigval, signal_handler, sigval, None)
    except NotImplementedError:
        for sigval in signals:
            signal.signal(sigval, signal_handler)

    return signal_event
